==> PyTSA-0.1.0-py3-none-any.whl: pytsa/preprocessing.py <==
'''
    author: rkopeinig
    script: Preprocessing
    description: Bunch of pre-processing algorithms.
                 See Proba-V Time-Series Analysis Toolbox Poster.
    version: 0.1
'''


import numpy as np
import pandas as pd
import collections
from itertools import count
from math import sqrt
from sklearn.metrics import mean_squared_error


def anomaly_detection(series, window_size, sigma=1.0):
    """Anomaly detection using rolling window.

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param window_size: Rolling window size
        :type window_size: integer

        :param sigma: Sigma
        :type sigma: integer

        :return: Series of anomalies of Time-Series
        :rtype: pandas.core.series.Series
    """
    window = np.ones(int(window_size)) / float(window_size)
    avg = np.convolve(series, window, 'same').tolist()

    residual = series - avg
    std = np.std(residual)
    _dict = pd.DataFrame.from_dict({'anomalies_dict': collections.OrderedDict([(index, y_i) for
                                                                              index, y_i, avg_i in
                                                                              izip(count(), series, avg)
                                                                              if (y_i > avg_i + (sigma * std)) | (
                                                                                      y_i < avg_i - (sigma * std))])})
    return _dict['anomalies_dict']


def rolling_std(series,window):
    """Provides rolling window calculations of Time-Series

    :param series: First One-dimensional ndarray with axis labels including Time-Series.
    :type series: pandas.core.series.Series

    :param window: Size of moving window. Number of observations to calculate statistics
    :type window: integer

    :return: Rolling standard deviation for given Time Series
    :rtype: pandas.core.series.Series
    """
    rolling = series.rolling(window, center=True)
    return pd.Series(rolling.std())


def rolling_mean(series,window):
    """Provides rolling window calculations of Time-Series

    :param series: First One-dimensional ndarray with axis labels including Time-Series.
    :type series: pandas.core.series.Series

    :param window: Size of moving window. Number of observations to calculate statistics
    :type window: integer

    :return: Rolling mean for given Time Series
    :rtype: pandas.core.series.Series
    """
    rolling = series.rolling(window, center=True)
    return pd.Series(rolling.mean())



def EMWA(series, span):
    """Provides exponentially-weighted moving average

    :param series: First One-dimensional ndarray with axis labels including Time-Series.
    :type series: pandas.core.series.Series

    :param span: Specify decay in terms of halflife, math: alpha = 1 - exp(log(0.5) / halflife
    :type span: integer

    :return: exponentially-weighted moving average using input index
    :rtype: pandas.core.series.Series
    """
    ewma = pd.stats.moments.ewma
    fwd = ewma(series, span=span)
    bwd = ewma(series[::-1], span=span)
    c = np.vstack((fwd, bwd[::-1]))
    c = np.mean(c, axis=0)
    return pd.Series(c, index=series.index)


def outlier_analysis_using_filtered_median(series, threshold=3):
    """Returns outliers using median filtering (non-linear) technique.

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param threshold: Median-based threshold
        :type threshold: integer

        :return: Series of outliers of Time-Series
        :rtype: pandas.core.series.Series

    """
    signal = series.copy()
    difference = np.abs(signal - np.median(signal))
    median_difference = np.median(difference)
    if median_difference == 0:
        s = 0
    else:
        s = difference / float(median_difference)
    mask = s > threshold
    signal[mask] = np.median(signal)
    return series[(np.where(signal.values != series.values)[0])]


def persistence_model_forecast(series):
    """Returns outliers using filtered median from Pandas.

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param threshold: Median-based threshold
        :type threshold: integer

        :return: Series of outliers of Time-Series
        :rtype: pandas.core.series.Series

    """
    X = series.values
    train, test = X[0:-(len(series) / 8)], X[-(len(series) / 4):]
    len(X), len(train), len(test)
    history = [x for x in train]
    predictions = list()
    for i in range(len(test)):
        # make prediction
        predictions.append(history[-1])
        # observation
        history.append(test[i])
    rmse = sqrt(mean_squared_error(test, predictions))
    return [rmse, test, predictions]


######## NOT USED ########


def outlier_analysis_rolling_median(series):
    """Returns outliers using rolling median from Pandas.

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :return: Series of outliers of Time-Series
        :rtype: pandas.core.series.Series

    """
    from pandas import rolling_median
    out = rolling_median(series, window=3, center=True).fillna(method='bfill').fillna(method='ffill')
    difference = np.abs(series - out)
    outlier_2 = difference > 2
    outlier_3 = difference > 3
    return series[outlier_2], series[outlier_3]


def outlier_analysis_using_linear_regression(series):
    """Returns outliers using linear regression from Statsmodels.

    :param series: One-dimensional ndarray with axis labels including Time-Series.
    :type series: pandas.core.series.Series

    :return: List of outliers of Time-Series
    :rtype: list
    """
    x = np.where(series.values)[0]
    y = series.values
    regression = ols("data ~ x", data=dict(data=y, x=x)).fit()
    test = regression.outlier_test()
    outliers = ((x[i], y[i]) for i, t in enumerate(test.icol(2)) if t < 0.5)
    return list(outliers)

==> PyTSA-0.2.0-py3-none-any.whl: pytsa/preprocessing.py <==
'''
    author: rkopeinig
    script: Preprocessing
    description: Bunch of pre-processing algorithms.
                 See Proba-V Time-Series Analysis Toolbox Poster.
    version: 0.1
'''


import numpy as np
import pandas as pd
import collections
import scipy.signal
from itertools import izip, count
from math import sqrt
from numpy import fft
from sklearn.metrics import mean_squared_error



def anomaly_detection(series, window_size, sigma=1.0):
    """Anomaly detection using rolling window.

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param window_size: Rolling window size
        :type window_size: integer

        :param sigma: Sigma
        :type sigma: integer

        :return: Series of anomalies of Time-Series
        :rtype: pandas.core.series.Series
    """
    window = np.ones(int(window_size)) / float(window_size)
    avg = np.convolve(series, window, 'same').tolist()

    residual = series - avg
    std = np.std(residual)
    _dict = pd.DataFrame.from_dict({'anomalies_dict': collections.OrderedDict([(index, y_i) for
                                                                              index, y_i, avg_i in
                                                                              izip(count(), series, avg)
                                                                              if (y_i > avg_i + (sigma * std)) | (
                                                                                      y_i < avg_i - (sigma * std))])})
    return _dict['anomalies_dict']


def fourierExtrapolation(series, n_predict):
    """Fourier Extrapolation on given Time-Series.

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param n_predict: Number of predictions
        :type n_predict: integer

        :return: Fourier Extrapolation results of given Time-Series
        :rtype: numpy.ndarray
    """
    n = series.size
    n_harm = 5
    t = np.arange(0, n)
    p = np.polyfit(t, series, 1)
    x_notrend = series - p[0] * t
    x_freqdom = fft.fft(x_notrend)
    f = fft.fftfreq(n)
    indexes = list(range(n))
    indexes.sort(key=lambda i: np.absolute(f[i]))
    t = np.arange(0, n + n_predict)
    restored_sig = np.zeros(t.size)
    for i in indexes[:1 + n_harm * 2]:
        ampli = np.absolute(x_freqdom[i]) / n
        phase = np.angle(x_freqdom[i])
        restored_sig += ampli * np.cos(2 * np.pi * f[i] * t + phase)
    return restored_sig + p[0] * t


def rolling_min(series, window):
    """Provides rolling window calculations of Time-Series

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param window: Size of moving window. Number of observations to calculate statistics
        :type window: integer

        :return: Rolling minimum for given Time Series
        :rtype: pandas.core.series.Series
    """
    return series.rolling(window).min()


def rolling_mean(series, window):
    """Provides rolling window calculations of Time-Series

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param window: Size of moving window. Number of observations to calculate statistics
        :type window: integer

        :return: Rolling mean for given Time Series
        :rtype: pandas.core.series.Series
    """
    return series.rolling(window).mean()


def rolling_median(series, window):
    """Provides rolling window calculations of Time-Series

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param window: Size of moving window. Number of observations to calculate statistics
        :type window: integer

        :return: Rolling median for given Time Series
        :rtype: pandas.core.series.Series
    """
    return series.rolling(window).median()


def rolling_std(series, window):
    """Provides rolling window calculations of Time-Series

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param window: Size of moving window. Number of observations to calculate statistics
        :type window: integer

        :return: Rolling standard deviation for given Time Series
        :rtype: pandas.core.series.Series
    """
    return series.rolling(window).std()


def rolling_max(series, window):
    """Provides rolling window calculations of Time-Series

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param window: Size of moving window. Number of observations to calculate statistics
        :type window: integer

        :return: Rolling maximum for given Time Series
        :rtype: pandas.core.series.Series
    """
    return series.rolling(window).max()


def exponentially_weighted_moving_average(series, span):
    """Provides exponentially-weighted moving average

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param span: Specify decay in terms of halflife, math: alpha = 1 - exp(log(0.5) / halflife
        :type span: integer

        :return: exponentially-weighted moving average using input index
        :rtype: pandas.core.series.Series
    """
    ewma = pd.stats.moments.ewma
    fwd = ewma(series, span=span)
    bwd = ewma(series[::-1], span=span)
    c = np.vstack((fwd, bwd[::-1]))
    c = np.mean(c, axis=0)
    return pd.Series(c, index=series.index)


def outlier_analysis_using_filtered_median(series, threshold=3):
    """Returns outliers using median filtering (non-linear) technique.

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param threshold: Median-based threshold
        :type threshold: integer

        :return: Series of outliers of Time-Series
        :rtype: pandas.core.series.Series

    """
    signal = series.copy()
    difference = np.abs(signal - np.median(signal))
    median_difference = np.median(difference)
    if median_difference == 0:
        s = 0
    else:
        s = difference / float(median_difference)
    mask = s > threshold
    signal[mask] = np.median(signal)
    return series[(np.where(signal.values != series.values)[0])]


def persistence_model_forecast(series):
    """Returns outliers using filtered median from Pandas.

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :param threshold: Median-based threshold
        :type threshold: integer

        :return: Series of outliers of Time-Series
        :rtype: pandas.core.series.Series

    """

    X = series.values
    train, test = X[0:-(len(series) / 8)], X[-(len(series) / 4):]
    len(X), len(train), len(test)
    history = [x for x in train]
    predictions = list()
    for i in range(len(test)):
        # make prediction
        predictions.append(history[-1])
        # observation
        history.append(test[i])
    rmse = sqrt(mean_squared_error(test, predictions))
    return [rmse, test, predictions]


######## NOT USED ########


def outlier_analysis_rolling_median(series):
    """Returns outliers using rolling median from Pandas.

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :return: Series of outliers of Time-Series
        :rtype: pandas.core.series.Series

    """
    from pandas import rolling_median
    out = rolling_median(series, window=3, center=True).fillna(method='bfill').fillna(method='ffill')
    difference = np.abs(series - out)
    outlier_2 = difference > 2
    outlier_3 = difference > 3
    return series[outlier_2], series[outlier_3]


def outlier_analysis_using_linear_regression(series):
    """Returns outliers using linear regression from Statsmodels.

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :return: List of outliers of Time-Series
        :rtype: list
    """
    x = np.where(series.values)[0]
    y = series.values
    regression = ols("data ~ x", data=dict(data=y, x=x)).fit()
    test = regression.outlier_test()
    outliers = ((x[i], y[i]) for i, t in enumerate(test.icol(2)) if t < 0.5)
    return list(outliers)


def peaking_times(series):
    """Returns peaking times of Time-Series

        :param series: One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :return: Series of peaking times of Time-Series
        :rtype: pandas.core.series.Series
    """
    time_series = series.values
    peaks = scipy.signal.find_peaks_cwt(time_series,np.arange(1, len(time_series)/4))
    index=[]
    for i in peaks:
        index.append(series.index[i])
    return pd.Series(peaks,index=index)
